import rclpy
from rclpy.node import Node
from rclpy.parameter import Parameter
from rclpy.action import ActionServer
from rclpy.action import GoalResponse, CancelResponse
from rclpy.callback_groups import ReentrantCallbackGroup

from transitions import Machine
import yaml
import numpy as np
from movement_primitives.promp import ProMP
from src.interaction_utils.serialization import flatten_population, unflatten_population

from interaction_msgs.action import TaskEvaluation


class TaskNode(Node):
    def __init__(self):
        super().__init__('task_node')

        # Task Attributes
        self.number_of_processed_trajectories = 0
        self.goal_dict = {}
        self.eval_strategy = 'robot'
        self.seed = self.declare_parameter('seed', None)
        # ROS2 Interfaces
        self.task_callback_group = ReentrantCallbackGroup()

        # Topic

        # Service

        # Action
        self._goal = None
        self._task_action = ActionServer(self,
                                         TaskEvaluation,
                                         'interaction/task_action',
                                         goal_callback=self._task_goal_callback,
                                         cancel_callback=self._task_cancel_callback,
                                         execute_callback=self._task_execute_callback,
                                         callback_group=self.task_callback_group)

        # State Machine
        self.state = None
        # States
        self.states = [
            'waiting_for_task_specs',
            'processing_non_interactive_input',
            'processing_interactive_input',
            'waiting_for_robot_response',
            'waiting_for_objective_function_response',
            'sending_request'
            'error_recovery'
        ]

        # initialize state machine
        self.machine = Machine(self, states=self.states, initial='waiting_for_task_specs')

        self.machine.add_transition(trigger='non_interactive_specs_received', source='waiting_for_task_specs', dest='processing_non_interactive_input')
        self.machine.add_transition(trigger='interactive_specs_received', source='waiting_for_task_specs', dest='processing_interactive_input')
        self.machine.add_transition(trigger='non_interactive_to_robot', source='processing_non_interactive_input', dest='waiting_for_robot_response')
        self.machine.add_transition(trigger='non_interactive_to_obj_fun', source='processing_non_interactive_input', dest='waiting_for_objective_function_response')
        self.machine.add_transition(trigger='interactive_to_robot', source='processing_interactive_input', dest='waiting_for_robot_response')
        self.machine.add_transition(trigger='interactive_to_obj_fun', source='processing_interactive_input', dest='waiting_for_objective_function_response')
        self.machine.add_transition(trigger='sending_robot_results', source='waiting_for_robot_response', dest='sending_request')
        self.machine.add_transition(trigger='sending_obj_fun_results', source='waiting_for_obj_fun_response', dest='sending_request')
        self.machine.add_transition(trigger='sending_back', source='sending_request', dest='waiting_for_task_specs')
        self.machine.add_transition(trigger='error_trigger', source='*', dest='error_recovery')
        self.machine.add_transition(trigger='recovery_complete', source='error_recovery', dest='waiting_for_task_specs')

    def destroy(self):
        self._task_action.destroy()
        super().destroy_node()

    # Helper function
    def compute_weighted_mean(self, user_parameters, optimizer_mean, weight_parameter):
        # Check if the weight_parameter is between 0 and 1
        if not 0 <= weight_parameter <= 1:
            self.get_logger().error(f'Invalid weight parameter for weighted average: {weight_parameter}')
            self.error_trigger()

        # Compute the weighted average
        weighted_mean = (weight_parameter * user_parameters) + ((1 - weight_parameter) * optimizer_mean)
        return weighted_mean

    def compute_promp_trajectory(self, parameter_set,
                                 number_of_dimensions, number_of_parameters_per_dimensions,
                                 duration, number_of_time_steps):
        time = np.linspace(0, duration, number_of_time_steps)
        promp = ProMP(number_of_dimensions / 2, n_weights_per_dim=number_of_parameters_per_dimensions)
        trajectory = promp.trajectory_from_weights(time, parameter_set)

        return trajectory


    # State Methods
    def on_enter_processing_non_interactive_input(self):
        # Unflatten parameter array to get the parameter sets
        parameter_sets = unflatten_population(self._goal.parameter_array,
                                              self._goal.number_of_population,
                                              self._goal.number_of_dimensions,
                                              self._goal.number_of_parameters_per_dimensions)
        # Compute trajectories for each set
        trajectories = [self.compute_promp_trajectory(parameter_sets[i, :, :],
                                                      self._goal.number_of_population,
                                                      self._goal.number_of_parameters_per_dimensions,
                                                      self._goal.duation,
                                                      self._goal.number_of_time_steps
                                                      ) for i in range(parameter_sets.shape[0])]

        # send the trajectories to the robot or objective function
        if self.eval_strategy == 'robot':
            # TODO: Implement Action Interface for Robot Evaluation
            self.non_interactive_to_robot()

        elif self.eval_strategy == 'obj_fun':
            # TODO: Implement Action Interface for Objective Function Evaluation
            self.non_interactive_to_obj_fun()

        else:
            self.get_logger().error(f"Unknown evaluation strategy: '{self.eval_strategy}'")
            self.error_trigger()

    def on_enter_processing_interactive_input(self):
        weigthed_average = self.compute_weighted_mean(self._goal.user_parameters,
                                                      self._goal.current_optimizer_mean,
                                                      self._goal.weight_parameter)

        promp = ProMP(self._goal.number_of_dimensions / 2,
                      n_weights_per_dim=self._goal.number_of_parameters_per_dimensions)

        promp.from_weight_distribution(weigthed_average,
                                       self._goal.user_covariance_diag)

        if self._goal.conditional_points is not None:
            # TODO: Fix the Action to add condititional cov
            pass
        time = np.linspace(0, self._goal.duration, self._goal.number_of_time_steps)
        random_state = np.random.RandomState(self.seed.value)
        trajectories = promp.sample_trajectories(time, self._goal.number_of_population, random_state)

        # send the trajectories to the robot or objective function
        if self.eval_strategy == 'robot':
            # TODO: Implement Action Interface for Robot Evaluation
            self.interactive_to_robot()

        elif self.eval_strategy == 'obj_fun':
            # TODO: Implement Action Interface for Objective Function Evaluation
            self.interactive_to_obj_fun()

        else:
            self.get_logger().error(f"Unknown evaluation strategy: '{self.eval_strategy}'")
            self.error_trigger()

    def on_enter_waiting_for_robot_response(self):
        pass

    def on_enter_waiting_for_objective_function_response(self):
        pass

    # Callback functions
    def _task_goal_callback(self, goal):
        self._goal = goal

        if goal.user_input:
            self.interactive_specs_received()
        else:
            self.non_interactive_specs_received()

        return GoalResponse.ACCEPT

    def _task_cancel_callback(self, _):
        self.error_trigger()
        return CancelResponse.ACCEPT

    async def _task_execute_callback(self, goal_handle):
        feedback_msg = TaskEvaluation.Feedback()
        result_msg = TaskEvaluation.Result()

        # Feedback  Loop
        while not goal_handle.is_cancel_requested():
            # Send Feedback msg
            feedback_msg.current_state = self.state
            feedback_msg.processed_trajectories = self.number_of_processed_trajectories
            goal_handle.publish_feedback(feedback_msg)

            if self.state == 'sending_request':
                if self.goal_dict == {}:
                    self.error_trigger()

                result_msg.score = self.goal_dict['score']
                result_msg.new_means = self.goal_dict['new_means']
                break

        if goal_handle.is_cancel_requested():
            goal_handle.canceled()
            result_msg.score = -1
            return result_msg

        self.sending_back()
        goal_handle.succeed()
        return result_msg
